from copy import copy

import numpy as np

from metaworlds.core import Serializable
from metaworlds.envs import Step
from metaworlds.envs.mujoco import MujocoEnv
from metaworlds.misc import logger
from metaworlds.misc.overrides import overrides


def smooth_abs(x, param):
    return np.sqrt(np.square(x) + np.square(param)) - param


class HalfCheetahEnv(MujocoEnv, Serializable):

    FILE = 'half_cheetah.xml'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Always call Serializable constructor last
        Serializable.quick_init(self, locals())

    def get_current_obs(self):
        return np.concatenate([
            self.sim.data.qpos.flatten()[1:],
            self.sim.data.qvel.flat,
            self.get_body_com("torso").flat,
        ])

    def get_body_xmat(self, body_name):
        return self.data.get_body_xmat(body_name).reshape((3, 3))

    def get_body_com(self, body_name):
        return self.data.get_body_xpos(body_name)

    def step(self, action):
        # Copy action first to remove side effects
        action_copy = copy(action)

        self.forward_dynamics(action_copy)
        next_obs = self.get_current_obs()
        action_copy = np.clip(action_copy, *self.action_bounds)
        ctrl_cost = 1e-1 * 0.5 * np.sum(np.square(action_copy))
        assert action.all() == action_copy.all()
        run_cost = -1 * self.get_body_comvel("torso")[0]
        cost = ctrl_cost + run_cost
        reward = -cost
        done = False
        return Step(next_obs, reward, done)

    @overrides
    def log_diagnostics(self, paths):
        progs = [
            path["observations"][-1][-3] - path["observations"][0][-3]
            for path in paths
        ]
        logger.record_tabular('AverageForwardProgress', np.mean(progs))
        logger.record_tabular('MaxForwardProgress', np.max(progs))
        logger.record_tabular('MinForwardProgress', np.min(progs))
        logger.record_tabular('StdForwardProgress', np.std(progs))
